/* collatz iteration
*  the 3n+1 conjecture.
*
*       if f(n) is defined as
*
*       f(n) = (even(n)) ? n/2 : 3*n+1;
*
*       does n->f(n) converge to 1 for all n ?
*/

#include <stdio.h>
#include <math.h>
#include <values.h>

#define OK      0
#define ERROR   -1

main(argc, argv)
int argc;
char **argv;
{
FILE   *fp = NULL;
char   *fn = NULL;
int     start, cnt, k, min, max, n;
int     o_max = 0;
double  xmin, xmax, c_xmax;

if (argc < 3) {
    printf ("Usage %s start_number count\n", *argv);
    return 0;
    }

start = atoi(*++argv);
cnt   = atoi(*++argv);
fn    = *++argv;
if (fn && *fn ==  '-' && *(fn+1)== 'o') {
    fn = *++argv;
    if (fn == NULL || 0 == strlen(fn)) return ERROR;
    fp = fopen(fn, "w");
    if (fp == NULL) return ERROR;
    fprintf(fp,"       N                    Max value excursion length\n");
    fclose(fp);
    }

c_xmax = 0.0;

for (k = cnt; 0 < k--; start++) {
    n = collatz(start, &min, &max);
    if (n >= 0) xmax = max; /*       no need to float */
    else        n = x_collatz(start, &xmin, &xmax);

    if (xmax > c_xmax) {
        c_xmax = xmax;
        printf("%8d Max=%24.0f d=%4d\n", start, xmax, n);
        if (fn) c_append(fn, start, xmax, n);
        }
    else if (n > o_max) {
        o_max = n;
        printf("%8d Max=%24.0f d=%4d\n", start, xmax, n);
        if (fn) c_append(fn, start, xmax, n);
        }
    }
return OK;
}

collatz(int x, int *min, int *max)
{
int     cnt;
for (cnt=0, *min = 0X3FFFFFFF, *max = 0; x != 1; cnt++) {
    if (x < *min) *min = x;
    if (x > *max) *max = x;
    x = (x&1) ? 3*x+1 : x/2;
    if (x < 0) return -1;
    }
return cnt;
}

x_collatz(int x, double *xmin, double *xmax)
{
extern double  floor();
int     cnt;
double  z, z1;
z = x;
for (cnt=0, *xmin = MAXDOUBLE, *xmax = MINDOUBLE; z != 1.0; cnt++) {
    if (z < *xmin) *xmin = z;
    if (z > *xmax) *xmax = z;
    z1 = floor(z/2.0);
    z = (z != 2.0*z1) ? 3*z+1 : z1;
    }
return cnt;
}

c_append(char *file, int x, double xmax, int n)
{
FILE   *fp;
if (file == NULL || NULL == (fp = fopen(file, "a")))
    return ERROR;

fprintf(fp, "%8d Max=%24.0f d=%4d\n", x, xmax, n);
return fclose(fp);
}





