/* MANCHESTER DIGITAL SUMMER 1998 */
/* Compile this with Turbo-C command line:  tcc -ml scroll.c */

#include <stdio.h>
#include<dos.h>
#include<io.h>
#include<fcntl.h>

#define NULL 0L
#define OK   0
#define ERROR   (-1)

#define BREAK   0X03    /* CTRL-C */
#define CTRLZ   0X1A    /* CTRL-Z */

/*      Unsolicited keyboard input */
#define DOS_ATTN        (0X00FF & bdos(0X06,0XFF, 0))
#define BRAKE   BREAK == DOS_ATTN /* STOP ! */

#define V_NR    25
#define V_NC    80

#define N_CHARS 95
#define F_NR    24
#define F_NCMAX 53
#define F_WB    24
#define SR_NC  140
#define FONT_NBYTES     F_NR * F_NCMAX
#define FX_SIZE sizeof(struct h_font)+FONT_NBYTES
#define FMAP(X) (((BYTE*)(X)) + sizeof(struct h_font))
#define SR_SIZE F_NR * SR_NC

#define IRAND(X)        (rand()%(X))

typedef unsigned char BYTE;     /* unsigned 1 byte data */
struct  h_font {int nr; int nc;};
typedef struct h_font *FONT;
FONT    alpha[96];
BYTE    shift_mat[F_NR * SR_NC];
int     h_rule[F_NCMAX];
unsigned long int    n_wait = 0;
unsigned int    n_erode = 0;
unsigned int    qflip = 0;

#define FLIP_SCREEN     0X01    /* flip screen */
#define FLIP_GLYPH      0X02    /* flip glyph */

BYTE   *vram = (BYTE*) 0XB8000000L;     /* video ram */

char *version_string="scroll.c -D4- 02/09/1998 11:59:11";


char    *usage_text[] = {
    "f=file name            A raw data file",
    "l=16-character string  Enclose this argument in quotes",
    "j=lut_index            Select a pre-stored table[j] (0-5)",
    "s=source file          Any file name",
    "d=delay                Short wait time (0-32000)",
    "e=erosion_factor       Integer (0-800)",
    "a=attr                 Initial screen attribute (0-255)",
    "v=string               Verbose: print version string",
    "t=n#index_list         Cycle luts.  n = cycle count",
    "    index_list         sequence of hex digits",
    "p=string               Permute lut. permutation of 16 hex digits",
    "m=mtring               move direction + - ++ -+. Default +",
    NULL
    };

#define MAX_LUT 16
#define LUT_SIZE 16

char   *lut_tab[] = {
    " 123456789ABCDEF",         /* default */
    " WRQ80&3C7v+~^.N",
    " WQ8S0&3Coci;:.N",
    " .^~+v7C3&08QRWN",
    " .:;icoC3&0S8QWN",
    ".123456789ABCDEF",
    " W^Q+073C&v8~R.N",
    " 3.&^0~8+QvR7WCN",
    " WRQ 0 3 7 + ^.N",
    " W  8 & C v ~ .N",
    " WR 8  3 7v+ ^ N",
    " WRQ80 3 7v ~  N",
    " MPQ6o/EDiu=-:,N",
    " +.,;'|$/\=_().-",
    " ۰۰۰",
    " "
    };

main(argc, argv)
int argc;
char **argv;

{
extern  int     atoi();
extern  long int atol();
char   *hex_to_j16();
char   *exefile   = *argv;
char   *font_file = NULL;
char   *text_file = NULL;
char   *lut_perm = NULL;
char   *lut_cmap = NULL;
char   *lut = lut_tab[0];
char    c, *s, *t;
int     attr = -1;
int     jlut = -1;
int     qv = 0;         /* verbose */
int     n_cycle = 0;    /* wait time before new lut */
int     j;

for(argv++, argc--; 0 < argc--; argv++)
    switch(c = *(s = *argv)) {
    case 'F':   /* font */
    case 'f': font_file = s+2;
        break;
    case 'S':   /* source text */
    case 's': text_file = s+2;
        break;
    case 'L':   /* lookup table */
    case 'l': lut = s+2;
        break;
    case 'D':       /* delay */
    case 'd':  n_wait = atol(s+2);
        break;
    case 'E':       /* erode */
    case 'e':  n_erode = atoi(s+2);
        break;
    case 'A':       /* attribute or colour */
    case 'a':  attr = atoi(s+2);
        break;
    case 'J':       /* select lut */
    case 'j':  jlut = atoi(s+2);
               if (jlut < MAX_LUT) lut = lut_tab[jlut];
        break;
    case 'T':       /* Cycle lut */
    case 't':  for (t = s = s+2; *s && *s != '#'; s++);
               if (*s == 0) break;
               *s = 0;
               n_cycle = atoi(t);
               lut_cmap = hex_to_j16(s+1);
        break;
    case 'P':       /* Permute lut */
    case 'p':  if (16 == strlen(s+2)) lut_perm = hex_to_j16 (s+2);
        break;
    case 'v':       /* verbose    */
    case 'V':  qv = 1;
        break;
    case 'm':   /* movement */
    case 'M':   j = *(s+2);
                if (j == '-' || j== 'L' || j == 'l') qflip |= FLIP_SCREEN;
                j = *(s+3);
                if (j == '-' || j== 'L' || j == 'l') qflip |= FLIP_GLYPH;
    default: break;
    }

j = get_video_mode();
vram = (j == 7) ? (BYTE*) 0XB0000000L : (BYTE*) 0XB8000000L;

if (qv) {
    printf("version=%s\n", version_string);
    if (lut_perm != NULL) {
        printf("permute string = ");
        for (j = 0; j < strlen(lut_perm); j++)
            printf(" %2d", 0X0F & lut_perm[j]);
        printf("\n");
        }
    if (lut_cmap != NULL) {
        printf("cycle count = %d\n", n_cycle);
        printf("cycle string   = ");
        for (j = 0; j < strlen(lut_cmap); j++)
            printf(" %2d", 0X0F & lut_cmap[j]);
        printf("\n");
        }
    }

if (font_file == NULL || text_file == NULL || lut == NULL)
   usage(exefile, usage_text);

printf("font = %s lut = %s src = %s\n", font_file, lut, text_file);
if (get_font(alpha, font_file)) ex_error("bad font file");

/* dfont(alpha, lut); */
if (attr != -1) vdu_set_attr(attr);

ww_puts(" ", 0, 0, 80);
ww_puts(text_file, 0, 60, 20);

if (lut_cmap == NULL)
    scroll(alpha, lut, lut_perm, text_file);
else
    xscroll (alpha, n_cycle, lut_cmap, lut_perm, text_file);
}

usage (char *exe, char **txt)
{
printf("%s f=font l=lookup_table s=text_source\n", exe);
for (; NULL != *txt; txt++)
    printf("%s\n", *txt);
exit(0);
}

ex_error(char *str)
{
printf("error: %s\n", str);
exit(1);
}

char   *hex_to_j16(char *str)
{
extern  BYTE *malloc();
char   *dst, *s, *t;
int     nc = strlen(str);
int     c, j;

if (NULL == (dst = (char*) malloc(nc+1))) return NULL;
for (t = dst, s = str, j = nc; 0 < j--; s++, t++) {
    c = *s;
    *t = 16 + (9 * (c > '@'))+ (0X0F & c);
    }
dst[nc] = 0;
return dst;
}

get_font(FONT *abc, char *fn)
{
extern BYTE   *malloc();
FONT    fx_crunch();
FONT    fx;
int     fd;
BYTE   *cbuf;
int     i, cnt_op;
fd = open(fn, O_BINARY | O_RDONLY);
if (fd == ERROR) return ERROR;
cbuf = malloc(FONT_NBYTES);
if (cbuf == NULL) return ERROR;
for (i = 0; i < N_CHARS; i++) {
      cnt_op = read(fd, cbuf, FONT_NBYTES);
      fx = fx_crunch(i+32, cbuf);
      if (fx == NULL) ex_error("malloc(..font..)\n");
      abc[i] = fx;
      }
close(fd);
free(cbuf);
return OK;
}

FONT fx_crunch(int k, BYTE *src)
{
extern BYTE   *malloc();
FONT    fx;
BYTE   *s, *t;
int     nc, i, j;
int     drop_a, drop_b;

fx = (FONT) malloc (FX_SIZE);
if (fx == NULL) return NULL;
memset(FMAP(fx), 0, FONT_NBYTES);

fx->nr = F_NR;
fx->nc = F_NCMAX;
for(s = src, j = 0; j < F_NCMAX; h_rule[j++] = 0);
for(i = F_NR; 0 < i--; )
    for(j = 0; j < F_NCMAX; j++, s++)
        h_rule[j] |= (*s != 0);


if (k == ' ') {
    fx->nc = F_WB;
    }
else {          /* trim */
    for (drop_a = j = 0; j < F_NCMAX && 0 == h_rule[j]; j++)
        drop_a++;
    for (drop_b = 0, j = F_NCMAX-1; 0 == h_rule[j] && j > 0; j--)
        drop_b++;
    nc = F_NCMAX - (drop_a + drop_b);

    fx->nc = nc;
    for (t = FMAP(fx), s = src + drop_a, i = F_NR;
         0 < i--;
         s += F_NCMAX, t+= F_NCMAX) {
             if (qflip & FLIP_GLYPH)
                  reverse_memcpy(t, s, nc);
             else memcpy(t, s, nc);
         }
    }
return fx;
}

dfont(FONT *abc, char *lut)     /* print font */
{
FONT    fc;
int     nr, nc, ic, i, j;
BYTE   *t, *s;

for (ic = 0; ic < N_CHARS; ic++) {
    fc = abc[ic];
    nr = fc->nr;
    nc = fc->nc;
    t  = FMAP(fc);
    printf("[C%4.4d] %c nr=%d nc=%d \n", ic, ic+32, nr, nc);
    for (i= F_NR; 0 < i--; t+= F_NCMAX) {
        for (s = t, j = nc; 0 < j--; s++) {
            putchar(lut[*s]);
            }
        putchar('\n');
        }
    }
return OK;
}

scroll(FONT *abc, char *lut, char *xp, char *src)
{
extern BYTE   *malloc();
FILE   *fp = fopen(src, "r");
BYTE   *vs;     /* virtual screen */
int     c = ' ';
int     ncols = SR_NC;
int     j, key;
char    lut_c[LUT_SIZE];

if (fp == NULL) return ERROR;
vs = malloc(SR_SIZE);
if (vs == NULL) return ERROR;
memset(vs, 0, SR_SIZE);
for (;;) {
    key = DOS_ATTN;
    if (key == CTRLZ) system("");       /* DOS prompt */
    else if (key != 0) break;
    s_wait(n_wait);
    if (n_erode) rs_zap(n_erode, 1);
    rs_map(vs, lut, 1);
    ncols = vs_sl(vs, ncols);
    if (ncols <= V_NC) {
        c = fgetc(fp);
        if (c == EOF) break;
        if (c < ' ' || c > '~') c = ' ';
        ncols = vs_catc(vs, ncols, abc[c-' ']);
        }
    else delay(1);

    if (xp != NULL) {   /* permute the lut */
        for (j = 0; j < LUT_SIZE; j++) lut_c[j] = lut[0X0F & xp[j]];
        memcpy(lut, lut_c, LUT_SIZE);
        }
    }
return OK;
}

xscroll(FONT *abc, int cnt_max, char *idx, char *xp, char *src)
{
extern BYTE   *malloc();
FILE   *fp = fopen(src, "r");
BYTE   *vs;     /* virtual screen */
char   *s_lut;
int     c = ' ';
int     ncols = SR_NC;
int     key;
int     ix = 0;
int     maxi = strlen(idx);
int     cnt  = 0;
int     j;
char    lut[LUT_SIZE];
char    lut_c[LUT_SIZE];

if (fp == NULL) return ERROR;
vs = malloc(SR_SIZE);
if (vs == NULL) return ERROR;
memset(vs, 0, SR_SIZE);

for (;;) {
    key = DOS_ATTN;
    if (key == CTRLZ) system("");       /* DOS prompt */
    else if (key != 0) break;
    s_wait(n_wait);
    if (cnt == 0) {    /* cycle through subset of lut_tab */
        s_lut = lut_tab[0X0F & idx[ix++]];
        memcpy(lut, s_lut, LUT_SIZE);
        if (ix == maxi) ix = 0;
        cnt = cnt_max;
        }
    else cnt--;
    if (n_erode) rs_zap(n_erode, 1);
    rs_map(vs, lut, 1);

    if (xp != NULL) {   /* permute the lut */
        for (j = 0; j < LUT_SIZE; j++) lut_c[j] = lut[0X0F & xp[j]];
        memcpy(lut, lut_c, LUT_SIZE);
        }

    ncols = vs_sl(vs, ncols);
    if (ncols <= V_NC) {
        c = fgetc(fp);
        if (c == EOF) break;
        if (c < ' ' || c > '~') c = ' ';
        ncols = vs_catc(vs, ncols, abc[c-' ']);
        }
    else delay(1);
    }
return OK;
}

vs_sl(BYTE *xx, int nc) /* shift left */
{
BYTE   *s, *t;
int     i, j;
nc--;
for (s = xx, i = F_NR; 0 < i--; s += SR_NC)
    for (t = s, j = nc; 0 < j--; t++)
        *t = *(t+1);
return nc;
}

vs_catc(BYTE *sr, int x, FONT fc)       /* add fc to shift register */
{
BYTE   *t;
BYTE   *s = FMAP(fc);
int     nc = fc->nc;
int     i;

for (t = sr + x, i = F_NR; 0 < i--; s += F_NCMAX, t += SR_NC) {
    *t = *(t+1) = 0;
    memcpy(t+2, s, nc);
    }
return x + 1 + nc;
}

rs_map(BYTE *src, BYTE* idx, int y) /* map shift register to screen */
{
BYTE   *t = vram + y * 2 * V_NC;
BYTE   *s, *tz;
int     i, j;
if (qflip & FLIP_SCREEN) {
        for (i = F_NR; 0 < i--; src += SR_NC, t += 2*V_NC)
            for (j = V_NC, tz = t + 2*V_NC-2, s = src; 0 < j--; s++, tz -=2)
                *tz = idx[*s];
        return OK;
        }
for (i = F_NR; 0 < i--; src += SR_NC)
    for (j = V_NC, s = src; 0 < j--; s++, t+=2) /* skip attribute byte */
        *t = idx[*s];
return OK;
}

ww_puts(char *src, int y, int x, int nc)
{
BYTE   *t = vram + y * 2 * V_NC + (x*2);
int     maxj = strlen(src);
char   *s;
int     j;

for (s = src, j = 0; j < maxj && j < nc; j++, s++, t+=2) *t = *s;
for (; j < nc; j++, t+=2) *t = ' ';
return OK;
}

vdu_set_attr(int si)    /* colour */
{
BYTE   *t;
int     j;
for (t=vram+1, j = V_NR*V_NC; 0<j--; t+=2) *t = si;
return OK;
}

s_wait(unsigned long cnt)        /* short wait */
{
while(cnt > 0) cnt--;
return cnt;
}

rs_zap(int np, int y) /* blank part of screen */
{
BYTE   *t = vram + y * 2 * V_NC;
BYTE   *s;
while (0 < np--) {
    s = t + 2 * IRAND(F_NR*V_NC);
    *s = ' ';
    }
return OK;
}

int get_video_mode()
{
union   REGS xx;
xx.h.ah = 0X0F;         /* set video mode */
xx.h.al = 0;
xx.x.bx = xx.x.cx = 0;
int86(0X10, &xx, &xx);
return (0X7F & xx.h.al);
}

reverse_memcpy(t, s, nc)        /* mirror movement option */
BYTE   *t;      /* destination */
BYTE   *s;      /* source */
int     nc;     /* count */
{
for (s += nc-1; 0 < nc--; *t++ = *s--);
return OK;
}

