/*
*       fx_zgets()      keyboard mouse
*                       can use to run quiz program
*       fx_inwin()      restricted input
*/

#include <htext.h>
#include <na.h>
#include <scrnio.h>

#define ZG_CU   0       /* move cursor here */
#define ZG_XA   1       /* invert attributes */
#define ZG_KS   4       /* exit, for further selection */

fx_zgets(dst, lo, hi)
MATRIX  dst;    /* index locator */
MATRIX  lo;     /* optional start position */
MATRIX  hi;     /* (1) list of y, x values or (2) window list */
{
BYTE   *s  = hi->ADDR;
int     n  = MSIZE(s);
int     nc = MCOLS(s);
int     j  = 0; /* starting position */
int     rc = 0;
int    *ix = INT_PTR(hi);
int     fc, c, nr, x, y;
int    *si;
struct SDESC tt;

if  (nc == 2 || nc == 4) nr = n / nc;
else return ERROR;
fc = (nc == 2) ? ZG_CU : ZG_XA;

if (lo != NULL) {       /* starting point */
    j = *(INT_PTR(lo));
    if (j < 0) j = 0;
    else if (j >= nr) j = nr-1;
    }

while (rc == 0) {       /* deal with fields */
    si = ix + nc * j;
    y = si[0]; x = si[1];
    switch(fc) {
    default:
    case ZG_CU: /* cursor */
        tcur(si[1], si[0]);
        c = vgetch();
        break;
    case ZG_XA: /* window */
        mm_xattr(si, 0);    /* invert attributes */
        c = vgetch();
        mm_xattr(si, 0);    /* invert attributes */
        break;
            }
    if (nr == 1) rc = 1;
    else
    switch(c) {         /* exit key */
    case CR:  rc = 1;
        break;
    case ESC:
    default: rc = ERROR;
        break;
    case KHOME: j = 0;
        break;
    case KEND:  j = nr-1;
        break;
    case KLEFT: /* previous field */
        if (--j < 0) j = nr-1;
        break;
    case HTAB:
    case KRGHT: /* next field */
        if (++j == nr) j = 0;
        break;
    case KUP:       /* nearest y-1, x */
         for (; si[0] >= y && si > ix; si -= nc);
         for (;si > ix && si[1] > x; si -= nc);
         j = (si-ix)/nc;
         break;
    case KDOWN:     /* nearest y+1, x */
         for (; si[0] <= y && si < ix + n-nc; si+=nc);
         for (;si < ix + n-nc && si[1] < x; si+=nc);
         j = (si-ix)/nc;
         break;
         }
    }
vkpush(c);
if (rc == ERROR) j = rc;
return mkik(dst, j);
}

fx_inwin(dst, src)      /* restrict input to 1-line window */
MATRIX  dst;            /* text */
MATRIX  src;            /* where */
{
BYTE   *s;
int     rc;
int     n, nr, nc;
int    *wx;
if (src == NULL || NULL == (s = src->ADDR) || TYPN != MTYPE(s))
        return ERROR;
n = rho(src, &nr, &nc);
if (nc < 4) return ERROR;
ww_mget(dst, src);              /* starting data */
wx = INT_PTR(src);
rc = ww_edlin(wx, dst);         /* edit single line */
vkpush(rc);                     /* keep key */
return OK;
}

#ifdef  TC

fx_inkey(dst, src)              /* keyboard input */
MATRIX  dst;
MATRIX  src;                    /* delay time to poll */
{
int    *si = INT_PTR(src);
int     n, k = 0;
long    int     cnt = 1;
for (n = MSIZE(src->ADDR); 0 < n--; cnt *= (long) (*si++));
while (0 < cnt--)  /* check keyboard */
    if (kbhit()) {
        k = vgetch();
        break;
        }
return  mkik(dst, k);
}

#else           /* 32-bit version, so long loop */

fx_inkey(dst, src)              /* keyboard input */
MATRIX  dst;
MATRIX  src;                    /* delay time to poll */
{
int    *si = INT_PTR(src);
int     n = si[0];
int     k = 0;
while (0 < n--)  /* check keyboard */
    if (kbhit()) {
        k = vgetch();
        break;
        }
return  mkik(dst, k);
}

#endif

fx_zputs(dst, data, idx)
MATRIX  dst;
MATRIX  data;
MATRIX  idx;
{
BYTE   *s  = idx->ADDR;
int     n  = MSIZE(s);
int     nc = MCOLS(s);
int     ns = MSIZE(data->ADDR);
int    *ix = INT_PTR(idx);
BYTE   *t;
int     nr;

if (n == 0 || nc == 0) return mkik(dst, 0);
nr = n / nc;
if (ns != 1 && ns != nr) return ERROR;
for (s = TEXT_PTR(data); 0 < nr--; ix += nc) {
     if (   ix[0] >= 0 && ix[0] < WW_MAXROWS
         && ix[1] >= 0 && ix[1] < WW_MAXCOLS) {
        t = WW_MMAP(ix[0], ix[1]);
        *t = *s;
        if (nc == 3) *(t+1) = ix[2];    /* colour */
        }
    if (ns > 1) s++;
    }

#ifdef  DJGPP
ScreenUpdate(mm_vptr);
#endif

return  mkik(dst, 0);
}

fx_getscr(dst, idx)
MATRIX  dst;
MATRIX  idx;
{
BYTE   *s  = idx->ADDR;
int     n  = MSIZE(s);
int     nc = MCOLS(s);
int    *ix = INT_PTR(idx);
BYTE   *t;
int     nr;

if (n == 0 || nc == 0) return mkik(dst, 0);
nr = n/nc;
if (nc != 2 || mkobj(dst, 'T', nr, nr)) return ERROR;
for (t = TEXT_PTR(dst); 0 < nr--; ix += nc) {
     if (   ix[0] >= 0 && ix[0] < WW_MAXROWS
         && ix[1] >= 0 && ix[1] < WW_MAXCOLS) {
        s = WW_MMAP(ix[0], ix[1]);
       *t++ = *s;
        }
    }
return  OK;
}
