/* WWMAP.C
*
*       Memory mapped screen IO functions
*/

#include <htext.h>
#include <na.h>
#define SCRMAP_DATA
#include <scrnio.h>
#include <dos.h>

#ifdef  DJGPP
#include <dt.h>
#include <pc.h>
#endif


fx_screen(dst, src)     /* enquire | set screen size */
MATRIX dst;             /* rows, columns             */
MATRIX src;             /* set with these values     */
{
BYTE   *s;
int    *ix;
if (src != NULL) {      /* set */
    s = src->ADDR;
    if (MTYPE(s) != TYPN || MSIZE(s) < 2) return ERROR;
    ix = INT_PTR(src);
    if (ix[0] < 0 || ix[0] > SCR_MAX_YVAL ||
        ix[1] < 0 || ix[1] > SCR_MAX_XVAL) return ERROR;
    ed_scr[2] = ix[0];
    ed_scr[3] = ix[1];

#ifdef DJGPP
    {
    MATRIX p;
    p = NA_PTR(pm_scr);
    if (mkobj(p, 'T', 2*ed_scr[2]*ed_scr[3], 2*ed_scr[3])) return ERROR;
    mm_vptr = TEXT_PTR(p);
    }

#endif
    }
if (mkobj(dst, 'N', 2, 2)) return ERROR;
ix = INT_PTR(dst);      /* values */
ix[0] = ed_scr[2];
ix[1] = ed_scr[3];
return OK;
}

#ifdef TC

fx_vram(dst, src)       /* enquire | set video ram pointer */
MATRIX dst;             /* for dos -- integer SEG, OFFSET  */
MATRIX src;             /* segment, offset in DOS          */
{
int     qget = (src == NULL);
BYTE   *s;
unsigned int    *ix;
if (mkobj(dst, 'N', 2, 2)) return ERROR;
ix = (unsigned *)INT_PTR(dst);  /* save old value */
ix[0] = FP_SEG(mm_vptr);
ix[1] = FP_OFF(mm_vptr);
if (qget) return OK;
/*      check input data        */
if (NULL == (s = src->ADDR) ||
    TYPN != MTYPE(s) ||
       2 != MSIZE(s))
        return ERROR;   /* invalid data */
ix = (unsigned *) INT_PTR(src);
FP_SEG(mm_vptr) = ix[0];
FP_OFF(mm_vptr) = ix[1];
return OK;
}

#else   /* DJGPP */

fx_vram(dst, src)       /* enquire | set video ram pointer */
MATRIX dst;             /* for dos -- integer SEG, OFFSET  */
MATRIX src;             /* segment, offset in DOS          */
{
if (OK== mkobj(dst, 'N', 1, 1)) {
   *(INT_PTR(dst)) = (int) mm_vptr;
    return OK;
    }
else return ERROR;
}

#endif

ww_menu(wvec, txt)      /* menu function */
MATRIX wvec;            /* window spec */
MATRIX txt;             /* menu data */
{
BYTE   *s  = txt->ADDR;
int    *wx = INT_PTR(wvec);
int     dy = wx[2];
int     qscroll = 0;
int     qcur = (MSIZE(wvec->ADDR) > 4);
int     rc, c, k;
int     nr, nc;
int     mx, my;
if (s == NULL || TYPT != MTYPE(s) || ww_size (wvec)) return ERROR;
k = rho(txt, &nr, &nc);
if (qcur) {             /* starting position */
        k = wx[4];
        if (k >= 0 && k < nr) MINDX(s) = k;
        }
k = MINDX(s);
if (dy < k) {
        MINDX(s) = (k / dy) * dy;
        k = k % dy;
        }
else    MINDX(s) = 0;

ww_mput(txt, wvec);
for    (rc = 0; rc == 0;) {
        qscroll = 0;
        mm_xattr(wx, k);
        c = vgetch();
        mm_xattr(wx, k);
        if (XMOUSE_KEY(c)) {    /* translate mouse coordinates to key */
            mx = XMOUSE_CX(c)-33;
            my = XMOUSE_CY(c)-33;
            if (mx < wx[1]-1  || mx > wx[1]+wx[3] ||
                my < wx[0]-1  || my > wx[0]+wx[2])
                c = ESC;
            else if (my == wx[0]-1)
                c = (mx == wx[1]-1)     ? KHOME: KPGUP;
            else if (my == wx[0]+wx[2])
                c = (mx == wx[1]+wx[3]) ? KEND : KPGDN;
            else {
                c = CR;
                k = my-wx[0];
                }
            }
        switch(c) {
        case CR:
                rc = c;
                break;
        case ESC:
        case KLEFT:
        case KRGHT:
                rc = ERROR;
                break;
        case KHOME:
        case KPGDN:
        case KPGUP:
        case KEND:
                qscroll = mm_scroll(c, txt, wx);
                break;
        case KUP:
                if (k > 0) k--;
                else qscroll = mm_scroll(c, txt, wx);
                break;
        case KDOWN:
        default:
                if (k < dy -1 ) k++;
                else qscroll = mm_scroll(c, txt, wx);
                break;
                }
        if (qscroll) ww_mput(txt, wvec);
        }
if (rc == ERROR) return rc;
MINDX(s) += k;
return MINDX(s);
}

mm_scroll(fc, xx, wx)   /* change xx[index] */
int     fc;             /* function code */
MATRIX  xx;             /* text */
int    *wx;             /* scroll region */
{
BYTE   *v;
int     k;
int     maxy;           /* window height */
int     n, nr, nc;
if (xx == NULL || NULL == (v = xx->ADDR)) return 0;
n = rho(xx, &nr, &nc);
k = MINDX(v);
maxy = wx[2];
switch (fc) {
        case KPGDN:     k += maxy;
                break;
        case KPGUP:     k -= maxy;
                break;
        case KUP:       k--;
                break;
        case KDOWN:     k++;
                break;
        case KHOME:     k = 0;
                break;
        case KEND:      k = nr - maxy;
                break;
        }
if (k < 0 || k + maxy > nr) return 0;
MINDX(v) = k;   /* update */
return 1;
}

mm_xattr(wx, idx)       /* invert attributes of line */
int *wx;        /* y0, x0, dy, dx for window */
int idx;        /* line offset */
{
BYTE   *dst;
BYTE    attr;
int     dx;
if (idx < 0 || idx >= wx[2])
        return(ERROR);
dst  = WW_MMAP(wx[0], wx[1]);   /* point y, x   */
dst +=  idx * V_2COLS;
for (dst++, dx = wx[3]; 0 < dx--; dst+=2) {
        attr = *dst;
        attr = (0X0F & (attr>>4)) | (0XF0 & (attr<<4));
        *dst = attr;
        }

#ifdef  DJGPP
ScreenUpdate(mm_vptr);
#endif

return(OK);
}

ww_cls()        /* clear screen */
{
int     n = 2 * WW_MAXROWS * WW_MAXCOLS;
BYTE   *t = WW_MMAP(0, 0);

while (0 < n--) {
    *t++ = ' ';         /* character */
    t++;                /* skip attribute */
    /* *t++ = SCR_NORM;   attribute */
    }

#ifdef  DJGPP
ScreenUpdate(mm_vptr);
#endif

return OK;
}

ww_vput(xx, ix, hcut, qflip)    /* write, and trim matrix to window */
MATRIX  xx;     /* data -- with cursor */
int    *ix;     /* window size */
int     hcut;   /* horizontal shift offset */
int     qflip;  /* reverse orientation */
{
BYTE   *v = xx->ADDR;
int     n = MSIZE(v);
int     nc = MCOLS(v);
BYTE   *t;
int    *iy;
int     i, j, nr, dy;
struct SDESC tt;

nr = (nc == 0) ? 0 : n / nc;
t = WW_MMAP(ix[0], ix[1]);
if (MTYPE(v) == TYPN) {         /* attributes */
    iy = INT_PTR(xx);
    dy = (n == ix[2]*ix[3]);
    for (i = ix[2]; 0 < i--; t += V_2COLS)
        for (v = t+1, j = ix[3]; 0 < j--; v++, iy += dy)
            *v++ = *iy;
    }
else if (MTYPE(v) == TYPT) {    /* text */
    tt.ADDR  = TEXT_PTR(xx) + hcut;
    tt.SDLEN = (nc <= hcut) ? 0 : nc - hcut;
    if (nr > 1) {
        nr -= MINDX(v);
        tt.ADDR  += nc * MINDX(v);
        }
    for (dy = ix[2]; 0 < dy-- ; nr--, t += V_2COLS, tt.ADDR += nc) {
        if (nr <= 0) tt.SDLEN = 0;
        if (qflip) mm_rlin(t, ix[3], &tt);
        else       mm_dlin(t, ix[3], &tt);
        }
  }

#ifdef  DJGPP
ScreenUpdate(mm_vptr);
#endif

return OK;
}

ww_mput(xx, wvec)       /* write, and trim matrix to window */
MATRIX xx;              /* data -- with cursor */
MATRIX wvec;            /* window spec */
{
return ww_size(wvec) || ww_vput(xx, INT_PTR(wvec), 0, 0);
}

ww_size(wvec)           /* validate window size */
MATRIX wvec;            /* Int vector Y, X, DY, DX */
{
BYTE *v;
int  *ix;
if (wvec == NULL || NULL == (v = wvec->ADDR) || TYPN != MTYPE(v))
    return ERROR;
ix = INT_PTR(wvec);
if     (ix[0] < 0  ||
        ix[1] < 0  ||
        ix[2] < 0  ||
        ix[3] < 0  ||
        ix[0] + ix[2] > WW_MAXROWS ||
        ix[1] + ix[3] > WW_MAXCOLS)
        return ERROR;
return OK;
}

mm_dlin(dst, max, ss)   /* paste a line */
BYTE *dst;              /* where on screen */
int max;                /* size */
struct SDESC *ss;       /* message */
{
BYTE *s = ss->ADDR;
int   n = ss->SDLEN;
while (0 < n-- && 0 < max--) {
    *dst++ = *s++;
     dst++;
     }
while (0 < max--) {
    *dst++ = ' ';
    *dst++;
    }

#ifdef  DJGPP
ScreenUpdate(mm_vptr);
#endif

return OK;
}

mm_rlin(dst, max, ss)   /* paste a reversed line */
BYTE *dst;              /* where on screen */
int max;                /* size */
struct SDESC *ss;       /* message */
{
BYTE *s = ss->ADDR;
int   n = ss->SDLEN;
dst += 2*(max-1);
while (0 < n-- && 0 < max--) {
    *dst-- = *s++;
     dst--;
     }
while (0 < max--) {
    *dst-- = ' ';
    *dst--;
    }

#ifdef  DJGPP
ScreenUpdate(mm_vptr);
#endif

return OK;
}

mmscr_sput(ww, ss)      /* paste a slice */
int  *ww;               /* where on screen */
SLICE ss;               /* message */
{
BYTE   *dst = WW_MMAP(ww[0], ww[1]);    /* point y, x   */
return mm_dlin(dst, ww[3], ss);
}

ww_vget(dst, wx)        /* read a box on the text screen */
MATRIX  dst;    /* data */
int    *wx;     /* window size (Y0,X0),(DY,DX) */
{
BYTE   *z = WW_MMAP(wx[0], wx[1]);
BYTE   *s, *t;
int     nr, nc;

#ifdef  DJGPP
ScreenRetrieve(mm_vptr);
#endif

if (mkobj(dst, 'T', wx[2]*wx[3], wx[3])) return ERROR;
for (t = TEXT_PTR(dst), nr = wx[2]; 0 < nr--; z += V_2COLS)
        for (s = z, nc = wx[3]; 0 < nc--; t++) {
        *t = *s++;      /* character */
        s++;            /* attribute */
        }
return OK;
}

ww_mget(dst, ww)        /* read screen */
MATRIX dst;             /* text matrix */
MATRIX ww;              /* window specifier */
{
return (dst == NULL ||
        ww_size(ww) ||
        ww_vget(dst, INT_PTR(ww)));
}
